/*
 *	(c) Copyright 1993 Paramax Systems Corp.
 *
 *	This material my be reproduced by or for the U.S.
 *	Government pursuant to the copyright license under
 *	the clause at DFARS 252.227-7013   (1991).
 */

#include <kqml/kqmlconf.h>

#include <sys/types.h>
#include <sys/socket.h>

#include <stdio.h>
#include <memory.h>
#include <unistd.h>
#include <errno.h>

#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifdef __KQML_MACH_SVR4__
#include <sys/systeminfo.h> /* sysinfo instead of gethostname */
#endif

#ifdef __KQML_MACH_SUN__
extern char *sys_errlist[];
#endif

#include <kqml.h>
#include <kqml/kqmllib.h>

static int close_fd (int);

#define digit(c) ((c) >= '0' && (c) <= '9')

#define MAX_EINTR_RETRY 10

/* CLIENT CONNECTIONS
 * Open a file descriptor to SERVICE at HOSTNAME.  If SERVICE is the
 * name of a service, then it must be known on the local machine.
 * SERVICE can also be the ASCII representation of a decimal number,
 * in which case it is interpreted as the port number to connect to.
 * Returns a valid file descriptor if successful, or a negative number
 * if not.  
 *
 *	-1 - can't resolve hostname
 *	-2 - can't determine what service port to connect to
 *	-3 - can't open a connection to the remote service
 */

int
_kqml_tcp_to_service (int port, char *hostname)
{
  struct sockaddr_in name;
  int connection = -1;
  int	res = -1;
  char localhost[256 + 1];
  unsigned long addr;
  struct hostent *host;
  int	attempts;

	/* find host address */
	if (! hostname)		/* if no host, use localhost */
	{
#		ifdef __KQML_MACH_SVR4__
		if (sysinfo (SI_HOSTNAME, localhost, sizeof (localhost)) == -1)
			hostname = "localhost";
#		else
		if (gethostname (localhost, sizeof (localhost)) == -1)
			hostname = "localhost";
#		endif
		else
			hostname = localhost;
	}

	addr = INADDR_BROADCAST;
	if (digit (*hostname))	/* if inet address provided */
	{
		addr = inet_addr (hostname);
	}

	if (addr == INADDR_BROADCAST)
	{
		host = gethostbyname (hostname);
	}
	else
	{
		host = gethostbyaddr ((char *)&addr, sizeof (addr), AF_INET);
	}

	if (host == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("tcp_to_serv: hostname: %s.", sys_errlist[errno]);
		return -1;
	}

	memset (&name, 0, sizeof (name));/* Prepare the socket name for binding. */
	name.sin_family = AF_INET;
	memcpy (&name.sin_addr, host->h_addr_list[0], host->h_length);

	name.sin_port = htons (port);

	/* Connect to the remote site */
	attempts = 0;
	do
	{
		connection = socket (AF_INET, SOCK_STREAM, IPPROTO_IP);
		if (connection > 0) 
		{
      	res = connect (connection, &name, sizeof (name));
      	if (res < 0)
			{
				_kqml_dbg_msg (KQML_DBG_ALWAYS) ("tcp_to_serv: connect: %s.", 
					sys_errlist[errno]);
	  			close_fd (connection);
			}
    	}
	}
	while (((res < 0) || (connection < 0))   /* still no connection */
		&& (errno == EINTR)		   /* and failure is not serious */
		&& (attempts++ < MAX_EINTR_RETRY) /* and haven't exceeded retry max */
	 );

	if ((res < 0) || (connection < 0))
	{
		if (connection > 0)
			close_fd (connection);
		return (-1);
	}

	return (connection);
}

static int
close_fd (int fd)
{
	int res;
	int count = 0;
  
	do
	{
		res = close (fd);
	}
	while (res && 
		(errno == EINTR) &&
		(count++ < MAX_EINTR_RETRY));

	return res;
}
