#include <kqml/kqmlconf.h>

#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>
#include <stdlib.h>

#include <kqml.h>
#include <kqml/kqmllib.h>

static void free_array (char wrapper_type, int count, void *ptr);

KQML_KMSG_LIST *
kqml_wrapper (KQML_REPLY_TYPE reply_type, void *reply)
{
	kqml_message *kmsg;
	int clen, wlen, count;
	char *wrapper_format, *content, *temp_format, *temp_content;

	/* If the reply is an error, and if reply is not NULL, it is
		the error message. Construct error performative and return 
		it 
	*/
	if (reply_type == KQML_REPLY_ERROR)
	{
		if (reply == NULL)
		{
			return KQML_HANDLER_ERROR;
		}
		else
		{
			return kqml_build_kmsg_list (NULL, 
				kqml_build_msg ("error", (char *)reply, NULL, NULL, NULL));
		}
	}
	else if (reply_type == KQML_REPLY_NONE)
	{
		return KQML_HANDLER_NONE;
	}
	else if (reply == NULL)
	{
		/* Is it an error, if reply_type is not KQML_REPLY_NONE and 
			reply == NULL ? */
		return KQML_HANDLER_ERROR;
	}
	else if (reply_type == KQML_REPLY_KMSG_LIST)
	{
		return (KQML_KMSG_LIST *) reply;
	}
	else if (reply_type == KQML_REPLY_KMSG)
	{
		return kqml_build_kmsg_list (NULL, (kqml_message *)reply);
	}
	else if (reply_type == KQML_REPLY_KMSG_ARRAY)
	{
		kqml_message **kmsg_ptr = (kqml_message **)reply;
		KQML_KMSG_LIST *kmsg_list;

		kmsg_list = KQML_HANDLER_NONE;
		while (*kmsg_ptr)
		{
			if (kqml_build_kmsg_list (&kmsg_list, *kmsg_ptr++) == 
				KQML_HANDLER_ERROR)
			{
				kqml_free_kmsg_list (kmsg_list);
				return KQML_HANDLER_ERROR;
			}
		}

		return kmsg_list;
	}
		
	/*
	 * wrapper format (for the raw results: ...)
	 * F:N:LEN1:LEN2:LEN3:...:LENN
	 * where F is the format (S-String, L-Long, B-Binary, U-Unsigned)
	 * N is the no of items
	 * LEN1 ... LENN is the length of each item
	 *		Optional: Only for string and binary
	*/ 

	switch (reply_type)
	{
		case KQML_REPLY_STRING:
		{
			char *string;

			content = string = (char *)reply;
			clen = strlen (string);

			wlen = 4 + kqml_digit_count (clen);
			temp_format = wrapper_format = malloc (wlen + 1);
			if (wrapper_format == NULL)
			{
				return KQML_HANDLER_ERROR;
			}
			wrapper_format[wlen] = '\0';

			*temp_format++ = 'S';
			*temp_format++ = ':';
			*temp_format++ = '1';
			*temp_format++ = ':';
			kqml_digit_to_str (clen, temp_format);

			kmsg = kqml_build_msg ("reply", content, NULL, NULL, 
				":kril-wrapper", wrapper_format, NULL);
			free (wrapper_format);
			if (kmsg == NULL)
			{
				return KQML_HANDLER_ERROR;
			}

			return kqml_build_kmsg_list (NULL, kmsg);
		}
		case KQML_REPLY_STRING_ARRAY:
		{
			int slen;
			char **temp, **string_ptr;

			string_ptr = (char **)reply;

			wlen = 1;	/* Format */
			clen = 0;
			count = 0;
			temp = string_ptr;
			do
			{
				slen = strlen (*temp);
				clen += slen + 1;
				wlen += kqml_digit_count (slen) + 1;
				count++;
			}
			while(*++temp);

			wlen += kqml_digit_count (count) + 1;

			temp_content = content = malloc (clen);
			if (content == NULL)
			{
				return KQML_HANDLER_ERROR;
			}

			temp_format = wrapper_format = malloc (wlen+1);
			if (wrapper_format == NULL)
			{
				free (content);
				return KQML_HANDLER_ERROR;
			}
			wrapper_format[wlen] = '\0';

			*temp_format++ = 'S';
			*temp_format++ = ':';
			temp_format += kqml_digit_to_str (count, temp_format);
			do
			{
				slen = strlen (*string_ptr);

				memcpy (temp_content, *string_ptr, slen);
				temp_content += slen;
				*temp_content++ = ' ';

				*temp_format++ = ':';
				temp_format += kqml_digit_to_str (slen, temp_format);
			}
			while (*++string_ptr);

			kmsg = kqml_build_msg ("reply", NULL, NULL, NULL, 
				":kril-wrapper", wrapper_format, NULL);
			free (wrapper_format);
			if (kmsg == NULL)
			{
				free (content);
				return KQML_HANDLER_ERROR;
			}

			if (kqml_put_field (":content", kmsg, content, clen) <= -1)
			{
				kqml_free_msg (kmsg);
				free (content);
				return KQML_HANDLER_ERROR;
			}
			else
			{
				free (content);
			}

			return kqml_build_kmsg_list (NULL, kmsg);
		}
		case KQML_REPLY_BIN:
		{
			KQML_BIN_CONTENT *bin_content = (KQML_BIN_CONTENT *)reply;

			clen = bin_content->len;
			content = bin_content->msg;

			wlen = 4 + kqml_digit_count (clen);
			temp_format = wrapper_format = malloc (wlen+1);
			if (wrapper_format == NULL)
			{
				return KQML_HANDLER_ERROR;
			}
			wrapper_format[wlen] = '\0';

			*temp_format++ = 'B';
			*temp_format++ = ':';
			*temp_format++ = '1';
			*temp_format++ = ':';
			kqml_digit_to_str (clen, temp_format);

			kmsg = kqml_build_msg ("reply", NULL, NULL, NULL, 
				":kril-wrapper", wrapper_format, NULL);
			free (wrapper_format);

			if (kmsg == NULL)
			{
				return KQML_HANDLER_ERROR;
			}

			if (kqml_put_field (":content", kmsg, content, clen) <= -1)
			{
				kqml_free_msg (kmsg);
				return KQML_HANDLER_ERROR;
			}

			return kqml_build_kmsg_list (NULL, kmsg);
		}
		case KQML_REPLY_BIN_ARRAY:
		{
			KQML_BIN_CONTENT **bin_content_ptr, **temp;

			temp = bin_content_ptr = (KQML_BIN_CONTENT **)reply;

			wlen = 1;
			clen = 0;
			count = 0;
			do
			{
				clen += (*temp)->len + 1;
				wlen += kqml_digit_count ((*temp)->len) + 1;
				count++;
			}
			while (*++temp);
			wlen += kqml_digit_count (count) + 1;

			temp_content = content = malloc (clen);
			if (content == NULL)
			{
				return KQML_HANDLER_ERROR;
			}

			temp_format = wrapper_format = malloc (wlen+1);
			if (wrapper_format == NULL)
			{
				free (content);
				return KQML_HANDLER_ERROR;
			}
			wrapper_format[wlen] = '\0';

			*temp_format++ = 'B';
			*temp_format++ = ':';
			temp_format += kqml_digit_to_str (count, temp_format);
			do
			{
				memcpy (temp_content, (*bin_content_ptr)->msg, 
					(*bin_content_ptr)->len);
				temp_content += (*bin_content_ptr)->len;
				*temp_content++ = ' ';

				*temp_format++ = ':';
				temp_format += kqml_digit_to_str ((*bin_content_ptr)->len, 
					temp_format);
			}
			while (*++bin_content_ptr);

			kmsg = kqml_build_msg ("reply", NULL, NULL, NULL, 
				":kril-wrapper", wrapper_format, NULL);
			free (wrapper_format);
			if (kmsg == NULL)
			{
				free (content);
				return KQML_HANDLER_ERROR;
			}

			if (kqml_put_field (":content", kmsg, content, clen) <= -1)
			{
				free (content);
				kqml_free_msg (kmsg);
				return KQML_HANDLER_ERROR;
			}
			else
			{
				free (content);
			}

			return kqml_build_kmsg_list (NULL, kmsg);
		}
		case KQML_REPLY_LONG:
		{
			long *long_value;

			long_value = (long *)reply;

			clen = kqml_digit_count (*long_value);
			content = malloc (clen);
			if (content == NULL)
			{
				return KQML_HANDLER_ERROR;
			}

			kqml_digit_to_str (*long_value, content);

			kmsg = kqml_build_msg ("reply", NULL, NULL, NULL, 
				":kril-wrapper", "L:1", NULL);
			if (kmsg == NULL)
			{
				free (content);
				return KQML_HANDLER_ERROR;
			}

			if (kqml_put_field (":content", kmsg, content, clen) <= -1)
			{
				free (content);
				kqml_free_msg (kmsg);
				return KQML_HANDLER_ERROR;
			}
			else
			{
				free (content);
			}

			return kqml_build_kmsg_list (NULL, kmsg);
		}
		case KQML_REPLY_LONG_ARRAY:
		{
			long **long_value_ptr, **temp;

			temp = long_value_ptr = (long **)reply;
			count = 0;
			wlen = 1;
			clen = 0;
			do
			{
				clen += kqml_digit_count (**temp) + 1;
				count++;
			}
			while (*++temp);
			wlen += kqml_digit_count (count) + 1;

			temp_content = content = malloc (clen);
			if (content == NULL)
			{
				return KQML_HANDLER_ERROR;
			}

			temp_format = wrapper_format = malloc (wlen+1);
			if (wrapper_format == NULL)
			{
				free (content);
				return KQML_HANDLER_ERROR;
			}
			wrapper_format[wlen] = '\0';

			*temp_format++ = 'L';
			*temp_format++ = ':';
			kqml_digit_to_str (count, temp_format);
			do
			{
				temp_content += kqml_digit_to_str (**long_value_ptr, temp_content);
				*temp_content++ = ' ';
			}
			while (*++long_value_ptr);

			kmsg = kqml_build_msg ("reply", NULL, NULL, NULL, 
				":kril-wrapper", wrapper_format, NULL);
			if (kmsg == NULL)
			{
				free (content);
				return KQML_HANDLER_ERROR;
			}

			if (kqml_put_field (":content", kmsg, content, clen) <= -1)
			{
				free (content);
				kqml_free_msg (kmsg);
				return KQML_HANDLER_ERROR;
			}
			else
			{
				free (content);
			}

			return kqml_build_kmsg_list (NULL, kmsg);
		}
		case KQML_REPLY_ULONG:
		{
			unsigned long *long_value;

			long_value = (unsigned long *)reply;
			clen = kqml_Udigit_count (*long_value);
			content = malloc (clen);
			if (content == NULL)
			{
				return KQML_HANDLER_ERROR;
			}
			kqml_Udigit_to_str (*long_value, content);

			kmsg = kqml_build_msg ("reply", NULL, NULL, NULL, 
				":kril-wrapper", "U:1", NULL);
			if (kmsg == NULL)
			{
				free (content);
				return KQML_HANDLER_ERROR;
			}

			if (kqml_put_field (":content", kmsg, content, clen) <= -1)
			{
				free (content);
				kqml_free_msg (kmsg);
				return KQML_HANDLER_ERROR;
			}
			else
			{
				free (content);
			}

			return kqml_build_kmsg_list (NULL, kmsg);
		}
		case KQML_REPLY_ULONG_ARRAY:
		{
			unsigned long **long_value_ptr, **temp;

			temp = long_value_ptr = (unsigned long **)reply;
			count = 0;
			wlen = 1;
			clen = 0;
			do
			{
				clen += kqml_Udigit_count (**temp) + 1;
				count++;
			}
			while (*++temp);
			wlen += kqml_digit_count (count) + 1;

			temp_content = content = malloc (clen);
			if (content == NULL)
			{
				return KQML_HANDLER_ERROR;
			}

			temp_format = wrapper_format = malloc (wlen+1);
			if (wrapper_format == NULL)
			{
				free (content);
				return KQML_HANDLER_ERROR;
			}
			wrapper_format[wlen] = '\0';

			*temp_format++ = 'U';
			*temp_format++ = ':';
			kqml_digit_to_str (count, temp_format);
			do
			{
				temp_content += kqml_Udigit_to_str (**long_value_ptr, 
					temp_content);
				*temp_content++ = ' ';
			}
			while (*++long_value_ptr);

			kmsg = kqml_build_msg ("reply", NULL, NULL, NULL, 
				":kril-wrapper", wrapper_format, NULL);
			free (wrapper_format);
			if (kmsg == NULL)
			{
				free (content);
				return KQML_HANDLER_ERROR;
			}

			if (kqml_put_field (":content", kmsg, content, clen) <= -1)
			{
				free (content);
				kqml_free_msg (kmsg);
				return KQML_HANDLER_ERROR;
			}
			else
			{
				free (content);
			}

			return kqml_build_kmsg_list (NULL, kmsg);
		}
		default:
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) 
				("kqml_wrapper: %d -- Unknown reply type\n", reply_type);
			return KQML_HANDLER_ERROR;
		}
	}
}

void *
kqml_unwrapper (kqml_message *kmsg, KQML_REPLY_TYPE *reply_type)
{
	int i, clen, wlen, count;
	char *wrapper_format, *content;
	char wrapper_type;

	wrapper_format = kqml_get_field (":kril-wrapper", kmsg, &wlen, NULL);
	content = kqml_get_field (":content", kmsg, &clen, NULL);
	if (wrapper_format == NULL || content == NULL)
	{
		*reply_type = KQML_REPLY_NONE;
		return NULL;
	}

	*reply_type = KQML_REPLY_ERROR;

	wrapper_type = wrapper_format[0];
	wrapper_format = strchr (wrapper_format, ':');
	if (wrapper_format == NULL || (count = atoi (++wrapper_format)) <= 0)
	{
		return NULL;
	}

	switch (wrapper_type)
	{
	case 'S':
		{
			int slen;
			char *string, **string_ptr;

			if (count == 1)
			{
				wrapper_format = strchr (wrapper_format, ':');
				if (wrapper_format == NULL || (slen = atoi (++wrapper_format)) < 0)
				{
					return NULL;
				}
				
				string = malloc (slen + 1);
				if (string == NULL)
				{
					return NULL;
				}

				memcpy (string, content, slen);
				string[slen] = '\0';

				*reply_type = KQML_REPLY_STRING;
				return (void *)string;
			}
			else
			{
				string_ptr =  malloc (sizeof (char *) * (count + 1));
				if (string_ptr == NULL)
				{
					return NULL;
				}
				
				for (i = 0; i < count; i++)
				{
					wrapper_format = strchr (wrapper_format, ':');
					if (wrapper_format == NULL || 
						(slen = atoi (++wrapper_format)) < 0)
					{
						free_array (wrapper_type, i-1, string_ptr);
						return NULL;
					}
				
					string = malloc (slen + 1);
					if (string == NULL)
					{
						free_array (wrapper_type, i, string_ptr);
						return NULL;
					}

					string_ptr[i] = string;
					memcpy (string, content, slen);
					string[slen] = '\0';
					content += slen + 1;
				}
			
				string_ptr[count] = NULL;
				
				*reply_type = KQML_REPLY_STRING_ARRAY;
				return string_ptr;
			}
		}
	case 'B':
		{
			int slen;
			KQML_BIN_CONTENT *bin_content, **bin_content_ptr;

			if (count == 1)
			{
				wrapper_format = strchr (wrapper_format, ':');
				if (wrapper_format == NULL || (slen = atoi (++wrapper_format)) < 0)
				{
					return NULL;
				}
				
				bin_content = malloc (sizeof (KQML_BIN_CONTENT));
				if (bin_content == NULL)
				{
					return NULL;
				}
				else
				{
					bin_content->len = slen;
					bin_content->msg = malloc (slen);
					if (bin_content->msg == NULL)
					{
						free (bin_content);
						return NULL;
					}
					memcpy (bin_content->msg, content, slen);
				}

				*reply_type = KQML_REPLY_BIN;
				return (void *)bin_content;
			}
			else
			{
				bin_content_ptr =  malloc (sizeof (KQML_BIN_CONTENT *) * 
					(count + 1));
				if (bin_content_ptr == NULL)
				{
					return NULL;
				}
				
				for (i = 0; i < count; i++)
				{
					wrapper_format = strchr (wrapper_format, ':');
					if (wrapper_format == NULL || 
						(slen = atoi (++wrapper_format)) < 0)
					{
						free_array (wrapper_type, i, bin_content_ptr);
						return NULL;
					}
				
					bin_content = malloc (sizeof (KQML_BIN_CONTENT));
					if (bin_content == NULL)
					{
						free_array (wrapper_type, i, bin_content_ptr);
						return NULL;
					}

					bin_content->len = slen;
					bin_content->msg = malloc (slen);
					if (bin_content->msg == NULL)
					{
						free (bin_content);
						free_array (wrapper_type, i, bin_content_ptr);
						return NULL;
					}
					bin_content_ptr[i] = bin_content;
					memcpy (bin_content->msg, content, slen);
					content += slen + 1;
				}
			
				bin_content_ptr[count] = NULL;
				
				*reply_type = KQML_REPLY_BIN_ARRAY;
				return bin_content_ptr;
			}
		}
	case 'L':
		{
			long *long_value, **long_value_ptr;
			char *end;

			if (count == 1)
			{
				long_value = malloc (sizeof (long));
				if (long_value == NULL)
				{
					return NULL;
				}

				/* I am not checking for conversion errors */
				*long_value = strtol (content, &end, 10);

				*reply_type = KQML_REPLY_LONG;
				return (void *)long_value;
			}
			else
			{
				long_value_ptr =  malloc (sizeof (long *) * (count + 1));
				if (long_value_ptr == NULL)
				{
					return NULL;
				}
				
				for (i = 0; i < count; i++)
				{
					long_value = malloc (sizeof (long));
					if (long_value == NULL)
					{
						free_array (wrapper_type, i, long_value_ptr);
						return NULL;
					}

					long_value_ptr[i] = long_value;
					/* I am not checking for conversion errors */
					*long_value = strtol (content, &end, 10);
					content = end + 1;
				}
			
				long_value_ptr[count] = NULL;
				
				*reply_type = KQML_REPLY_LONG_ARRAY;
				return long_value_ptr;
			}
		}
	case 'U':
		{
			unsigned long *long_value, **long_value_ptr;

			if (count == 1)
			{
				long_value = malloc (sizeof (unsigned long));
				if (long_value == NULL)
				{
					return NULL;
				}

				/* strtoul is not available in sunos. So use sscanf */
				sscanf (content, "%lu", long_value);

				*reply_type = KQML_REPLY_ULONG;
				return (void *)long_value;
			}
			else
			{
				long_value_ptr =  malloc (sizeof (unsigned long *) * (count + 1));
				if (long_value_ptr == NULL)
				{
					return NULL;
				}
				
				for (i = 0; i < count; i++)
				{
					long_value = malloc (sizeof (unsigned long));
					if (long_value == NULL)
					{
						free_array (wrapper_type, i, long_value_ptr);
						return NULL;
					}

					long_value_ptr[i] = long_value;
					/* strtoul is not available in sunos. So use sscanf */
					sscanf (content, "%lu", long_value);
					content = strchr (content, ' ');
					content++;
				}
			
				long_value_ptr[count] = NULL;
				
				*reply_type = KQML_REPLY_ULONG_ARRAY;
				return long_value_ptr;
			}
		}
	default:
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) 
				("kqml_unwrapper: %c -- Unknown wrapper type\n", wrapper_type);
			return KQML_HANDLER_ERROR;
		}
	}
}

static void
free_array (char wrapper_type, int count, void *ptr)
{
	int i;

	switch (wrapper_type)
	{
		case 'S':
			{
				char **string_ptr;

				string_ptr = (char **)ptr;
				for (i = 0; i < count; i++)
				{
					free (string_ptr[i]);
				}
				free (string_ptr);
				break;
			}
		case 'B':
			{
				KQML_BIN_CONTENT **bin_content_ptr;
				
				bin_content_ptr = (KQML_BIN_CONTENT **)ptr;
				for (i = 0; i < count; i++)
				{
					free (bin_content_ptr[i]);
				}
				free (bin_content_ptr);
				break;
			}
		case 'L':
		case 'U':
			{
				long **long_value_ptr;

				long_value_ptr = (long **)ptr;
				for (i = 0; i < count; i++)
				{
					free (long_value_ptr[i]);
				}
				free (long_value_ptr);
				break;
			}
	}

	return;
}

#ifdef TESTING
main ()
{
	kqml_message *kmsg, *kmsg_array[3];

	char *string = "This is a test string";
	char *string_array[3] = {
			"Small test string",
			"One more",
			NULL
		};

	long long_value = -2121969, 
		long_value1 = -21219691, long_value2 = -21219692;
	long *long_value_array[3] = { &long_value1, &long_value2, NULL };

	unsigned long ulong_value = 2121969, 
		ulong_value1 = 21219691, ulong_value2 = 21219692;
	unsigned long *ulong_value_array[3] = { &ulong_value1, &ulong_value2, NULL };

	KQML_BIN_CONTENT bin_content = { 22, "Message with a \" quote" },
		bin_content1 = { 24, "A message with a \" quote" },
		bin_content2 = { 23, "One more with a \" quote" };
	KQML_BIN_CONTENT *bin_content_array[3] = {
			&bin_content1, &bin_content2, NULL };

	void *reply;
	KQML_REPLY_TYPE reply_type;
	KQML_KMSG_LIST *kmsg_list;

	kmsg = kqml_build_msg ("test", "test content", NULL, NULL, NULL);
	kmsg_array[0] = kqml_build_msg ("test1", "test content 1", NULL, NULL, NULL);
	kmsg_array[1] = kqml_build_msg ("test2", "test content 2", NULL, NULL, NULL);
	kmsg_array[2] = NULL;

	kqml_print_kmsg_list (stderr, kqml_wrapper (KQML_REPLY_KMSG, kmsg));
	kqml_print_kmsg_list (stderr, kqml_wrapper (KQML_REPLY_KMSG_ARRAY, 
		kmsg_array));

	kmsg_list = kqml_wrapper (KQML_REPLY_STRING, string);
	kqml_print_kmsg_list (stderr, kmsg_list);
	reply = kqml_unwrapper (kmsg_list->kmsg, &reply_type);
	if (reply_type != KQML_REPLY_STRING)
	{
		fprintf (stderr, "Error: Unknown reply_type: %d\n", reply_type);
	}
	else
	{
		char *string = (char *)reply;

		fprintf (stderr, "Reply is %s\n", string);
		free (string);
	}

	kmsg_list = kqml_wrapper (KQML_REPLY_STRING_ARRAY, string_array);
	kqml_print_kmsg_list (stderr, kmsg_list);
	reply = kqml_unwrapper (kmsg_list->kmsg, &reply_type);
	if (reply_type != KQML_REPLY_STRING_ARRAY)
	{
		fprintf (stderr, "Error: Unknown reply_type: %d\n", reply_type);
	}
	else
	{
		char **temp_ptr, **string_ptr = (char **)reply;

		temp_ptr = string_ptr;
		do
		{
			fprintf (stderr, "Reply is %s\n", *temp_ptr);
		}
		while (*++temp_ptr);
		free_array ('S', 3, string_ptr);
	}

	kmsg_list = kqml_wrapper (KQML_REPLY_LONG, &long_value);
	kqml_print_kmsg_list (stderr, kmsg_list);
	reply = kqml_unwrapper (kmsg_list->kmsg, &reply_type);
	if (reply_type != KQML_REPLY_LONG)
	{
		fprintf (stderr, "Error: Unknown reply_type: %d\n", reply_type);
	}
	else
	{
		long *long_value = (long *)reply;

		fprintf (stderr, "Reply is %ld\n", *long_value);
		free (long_value);
	}

	kmsg_list = kqml_wrapper (KQML_REPLY_LONG_ARRAY, long_value_array);
	kqml_print_kmsg_list (stderr, kmsg_list);
	reply = kqml_unwrapper (kmsg_list->kmsg, &reply_type);
	if (reply_type != KQML_REPLY_LONG_ARRAY)
	{
		fprintf (stderr, "Error: Unknown reply_type: %d\n", reply_type);
	}
	else
	{
		long **temp_ptr, **long_value_ptr = (long **)reply;

		temp_ptr = long_value_ptr;
		do
		{
			fprintf (stderr, "Reply is %ld\n", **temp_ptr);
		}
		while (*++temp_ptr);
		free_array ('L', 3, long_value_ptr);
	}

	kmsg_list = kqml_wrapper (KQML_REPLY_ULONG, &ulong_value);
	kqml_print_kmsg_list (stderr, kmsg_list);
	reply = kqml_unwrapper (kmsg_list->kmsg, &reply_type);
	if (reply_type != KQML_REPLY_ULONG)
	{
		fprintf (stderr, "Error: Unknown reply_type: %d\n", reply_type);
	}
	else
	{
		unsigned long *long_value = (unsigned long *)reply;

		fprintf (stderr, "Reply is %lu\n", *long_value);
		free (long_value);
	}

	kmsg_list = kqml_wrapper (KQML_REPLY_ULONG_ARRAY, ulong_value_array);
	kqml_print_kmsg_list (stderr, kmsg_list);
	reply = kqml_unwrapper (kmsg_list->kmsg, &reply_type);
	if (reply_type != KQML_REPLY_ULONG_ARRAY)
	{
		fprintf (stderr, "Error: Unknown reply_type: %d\n", reply_type);
	}
	else
	{
		unsigned long **temp_ptr, **long_value_ptr = (unsigned long **)reply;

		temp_ptr = long_value_ptr;
		do
		{
			fprintf (stderr, "Reply is %lu\n", **temp_ptr);
		}
		while (*++temp_ptr);
		free_array ('U', 3, long_value_ptr);
	}

	kmsg_list = kqml_wrapper (KQML_REPLY_BIN, &bin_content);
	kqml_print_kmsg_list (stderr, kmsg_list);
	reply = kqml_unwrapper (kmsg_list->kmsg, &reply_type);
	if (reply_type != KQML_REPLY_BIN)
	{
		fprintf (stderr, "Error: Unknown reply_type: %d\n", reply_type);
	}
	else
	{
		KQML_BIN_CONTENT *bin_content = (KQML_BIN_CONTENT *)reply;

		fprintf (stderr, "Reply is %*.*s\n", bin_content->len,
			bin_content->len, bin_content->msg);
		free (bin_content->msg);
		free (bin_content);
	}

	kmsg_list = kqml_wrapper (KQML_REPLY_BIN_ARRAY, bin_content_array);
	kqml_print_kmsg_list (stderr, kmsg_list);
	reply = kqml_unwrapper (kmsg_list->kmsg, &reply_type);
	if (reply_type != KQML_REPLY_BIN_ARRAY)
	{
		fprintf (stderr, "Error: Unknown reply_type: %d\n", reply_type);
	}
	else
	{
		KQML_BIN_CONTENT **temp, **bin_content_ptr = (KQML_BIN_CONTENT **)reply;

		temp = bin_content_ptr;
		do
		{
			fprintf (stderr, "Reply is %*.*s\n", (*temp)->len,
				(*temp)->len, (*temp)->msg);
		}
		while (*++temp);
		free_array ('B', 3, bin_content_ptr);
	}
}
#endif
